/*  $Id: WeakHashtable.java 1035 2010-06-15 13:08:33Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.lang.ref.WeakReference;
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * A Weak Hashtable stores objects while letting garbage collector collect them.
 * Permits to use automatic memory management as a transparent cache.
 * @version $Id: WeakHashtable.java 1035 2010-06-15 13:08:33Z sdevaux $
 * @author Sebastien Devaux
 */
public class WeakHashtable<K,V> implements DataSource<K,V> {
    /**
     * Create new WeakHashtable.
     * @param source source where to search for uncached objects.
     */
	public WeakHashtable(DataSource<K,V> source) {
		_dataSource=source;
		_weakRefs=new Hashtable<K,WeakReference<V>>();
	}
    // Datasource implementation.
	public V get(K key) {
		V res=null;
		synchronized(this) {
			WeakReference<V> ref=_weakRefs.get(key);
			if (ref!=null) {
				res=ref.get();
			}	
			if (res==null) {
				res=_dataSource.get(key);
				if (res!=null) {
					_weakRefs.put(key,new WeakReference<V>(res));
					cleanKey();
				}
			}
		}	
		return res;
	}
    /**
     * Remove obsolete keys.
     * To avoid key collection overload, remove keys related to already
     * collected objects.
     */
	private void cleanKey() {
		_kcCount++;
		if (_kcCount>=DEF_KCCOUNT) {
			_kcCount=0;
			Enumeration<K> keys=_weakRefs.keys();
			while (keys.hasMoreElements()) {
				K key=keys.nextElement();
				WeakReference<V> ref=_weakRefs.get(key);
				if (ref.get()==null) {
					_weakRefs.remove(key);
				}
			}
		}	
	}
    /** weak reference table */
	private Hashtable<K,WeakReference<V>> _weakRefs;
    /** Data source to retrieve garbage collected objects */
	private DataSource<K,V> _dataSource;
    /** key clean count */
	private int _kcCount=0;
    /** null return count to reach to trigger key clean. */
	private static int DEF_KCCOUNT=256;
}
